@@ -22,11 +22,15 @@
 #ifndef __CONNMAN_SETTING_H
 #define __CONNMAN_SETTING_H
 
+#include <connman/types.h>
+
 #ifdef __cplusplus
 extern "C" {
 #endif
 
 connman_bool_t connman_setting_get_bool(const char *key);
+char **connman_setting_get_string_list(const char *key);
+unsigned int *connman_setting_get_uint_list(const char *key);
 
 #ifdef __cplusplus
 }
