@@ -49,6 +49,7 @@ struct connman_ipconfig {
 	const struct connman_ipconfig_ops *ops;
 	void *ops_data;
 
+	connman_bool_t enabled;
 	enum connman_ipconfig_method method;
 	struct connman_ipaddress *address;
 	struct connman_ipaddress *system;
@@ -506,6 +507,14 @@ void __connman_ipconfig_unset_rp_filter(int old_value)
 	connman_info("rp_filter restored to %d", old_value);
 }
 
+gboolean __connman_ipconfig_ipv6_privacy_enabled(struct connman_ipconfig *ipconfig)
+{
+	if (ipconfig == NULL)
+		return FALSE;
+
+	return ipconfig->ipv6_privacy_config == 0 ? FALSE : TRUE;
+}
+
 static void free_ipdevice(gpointer data)
 {
 	struct connman_ipdevice *ipdevice = data;
@@ -514,12 +523,12 @@ static void free_ipdevice(gpointer data)
 							ipdevice->index);
 
 	if (ipdevice->config_ipv4 != NULL) {
-		connman_ipconfig_unref(ipdevice->config_ipv4);
+		__connman_ipconfig_unref(ipdevice->config_ipv4);
 		ipdevice->config_ipv4 = NULL;
 	}
 
 	if (ipdevice->config_ipv6 != NULL) {
-		connman_ipconfig_unref(ipdevice->config_ipv6);
+		__connman_ipconfig_unref(ipdevice->config_ipv6);
 		ipdevice->config_ipv6 = NULL;
 	}
 
@@ -541,6 +550,13 @@ static void __connman_ipconfig_lower_up(struct connman_ipdevice *ipdevice)
 {
 	DBG("ipconfig ipv4 %p ipv6 %p", ipdevice->config_ipv4,
 					ipdevice->config_ipv6);
+
+	if (ipdevice->config_ipv6 != NULL &&
+			ipdevice->config_ipv6->enabled == TRUE)
+		return;
+
+	set_ipv6_state(ipdevice->ifname, FALSE);
+	ipdevice->ipv6_enabled = FALSE;
 }
 
 static void __connman_ipconfig_lower_down(struct connman_ipdevice *ipdevice)
@@ -575,9 +591,9 @@ static void update_stats(struct connman_ipdevice *ipdevice,
 		return;
 
 	if (ipdevice->config_ipv4)
-		service = connman_ipconfig_get_data(ipdevice->config_ipv4);
+		service = __connman_ipconfig_get_data(ipdevice->config_ipv4);
 	else if (ipdevice->config_ipv6)
-		service = connman_ipconfig_get_data(ipdevice->config_ipv6);
+		service = __connman_ipconfig_get_data(ipdevice->config_ipv6);
 	else
 		return;
 
@@ -798,6 +814,9 @@ void __connman_ipconfig_newaddr(int index, int family, const char *label,
 	connman_info("%s {add} address %s/%u label %s family %d",
 		ipdevice->ifname, address, prefixlen, label, family);
 
+	if (type == CONNMAN_IPCONFIG_TYPE_IPV4)
+		__connman_ippool_newaddr(index, address, prefixlen);
+
 	if (ipdevice->config_ipv4 != NULL && family == AF_INET)
 		connman_ipaddress_copy(ipdevice->config_ipv4->system,
 					ipaddress);
@@ -857,6 +876,9 @@ void __connman_ipconfig_deladdr(int index, int family, const char *label,
 	ipdevice->address_list = g_slist_remove(ipdevice->address_list,
 								ipaddress);
 
+	if (type == CONNMAN_IPCONFIG_TYPE_IPV4)
+		__connman_ippool_deladdr(index, address, prefixlen);
+
 	connman_ipaddress_clear(ipaddress);
 	g_free(ipaddress);
 
@@ -900,7 +922,6 @@ void __connman_ipconfig_newroute(int index, int family, unsigned char scope,
 
 	if (scope == 0 && (g_strcmp0(dst, "0.0.0.0") == 0 ||
 						g_strcmp0(dst, "::") == 0)) {
-		GSList *list;
 		GList *config_list;
 		enum connman_ipconfig_type type;
 
@@ -929,13 +950,6 @@ void __connman_ipconfig_newroute(int index, int family, unsigned char scope,
 		} else
 			return;
 
-		for (list = ipdevice->address_list; list; list = list->next) {
-			struct connman_ipaddress *ipaddress = list->data;
-
-			g_free(ipaddress->gateway);
-			ipaddress->gateway = g_strdup(gateway);
-		}
-
 		for (config_list = g_list_first(ipconfig_list); config_list;
 					config_list = g_list_next(config_list)) {
 			struct connman_ipconfig *ipconfig = config_list->data;
@@ -949,8 +963,8 @@ void __connman_ipconfig_newroute(int index, int family, unsigned char scope,
 			if (ipconfig->ops == NULL)
 				continue;
 
-			if (ipconfig->ops->ip_bound)
-				ipconfig->ops->ip_bound(ipconfig);
+			if (ipconfig->ops->route_set)
+				ipconfig->ops->route_set(ipconfig);
 		}
 	}
 
@@ -972,7 +986,6 @@ void __connman_ipconfig_delroute(int index, int family, unsigned char scope,
 
 	if (scope == 0 && (g_strcmp0(dst, "0.0.0.0") == 0 ||
 						g_strcmp0(dst, "::") == 0)) {
-		GSList *list;
 		GList *config_list;
 		enum connman_ipconfig_type type;
 
@@ -999,13 +1012,6 @@ void __connman_ipconfig_delroute(int index, int family, unsigned char scope,
 		} else
 			return;
 
-		for (list = ipdevice->address_list; list; list = list->next) {
-			struct connman_ipaddress *ipaddress = list->data;
-
-			g_free(ipaddress->gateway);
-			ipaddress->gateway = NULL;
-		}
-
 		for (config_list = g_list_first(ipconfig_list); config_list;
 					config_list = g_list_next(config_list)) {
 			struct connman_ipconfig *ipconfig = config_list->data;
@@ -1019,8 +1025,8 @@ void __connman_ipconfig_delroute(int index, int family, unsigned char scope,
 			if (ipconfig->ops == NULL)
 				continue;
 
-			if (ipconfig->ops->ip_release)
-				ipconfig->ops->ip_release(ipconfig);
+			if (ipconfig->ops->route_unset)
+				ipconfig->ops->route_unset(ipconfig);
 		}
 	}
 
@@ -1242,6 +1248,7 @@ static struct connman_ipconfig *create_ipv6config(int index)
 	ipv6config->refcount = 1;
 
 	ipv6config->index = index;
+	ipv6config->enabled = FALSE;
 	ipv6config->type = CONNMAN_IPCONFIG_TYPE_IPV6;
 	ipv6config->method = CONNMAN_IPCONFIG_METHOD_AUTO;
 	ipv6config->ipv6_privacy_config = 0;
@@ -1266,7 +1273,7 @@ static struct connman_ipconfig *create_ipv6config(int index)
  *
  * Returns: a newly-allocated #connman_ipconfig structure
  */
-struct connman_ipconfig *connman_ipconfig_create(int index,
+struct connman_ipconfig *__connman_ipconfig_create(int index,
 					enum connman_ipconfig_type type)
 {
 	struct connman_ipconfig *ipconfig;
@@ -1283,6 +1290,7 @@ struct connman_ipconfig *connman_ipconfig_create(int index,
 	ipconfig->refcount = 1;
 
 	ipconfig->index = index;
+	ipconfig->enabled = FALSE;
 	ipconfig->type = CONNMAN_IPCONFIG_TYPE_IPV4;
 
 	ipconfig->address = connman_ipaddress_alloc(AF_INET);
@@ -1305,9 +1313,12 @@ struct connman_ipconfig *connman_ipconfig_create(int index,
  *
  * Increase reference counter of ipconfig
  */
-struct connman_ipconfig *connman_ipconfig_ref(struct connman_ipconfig *ipconfig)
+struct connman_ipconfig *
+__connman_ipconfig_ref_debug(struct connman_ipconfig *ipconfig,
+				const char *file, int line, const char *caller)
 {
-	DBG("ipconfig %p refcount %d", ipconfig, ipconfig->refcount + 1);
+	DBG("%p ref %d by %s:%d:%s()", ipconfig, ipconfig->refcount + 1,
+		file, line, caller);
 
 	__sync_fetch_and_add(&ipconfig->refcount, 1);
 
@@ -1320,12 +1331,14 @@ struct connman_ipconfig *connman_ipconfig_ref(struct connman_ipconfig *ipconfig)
  *
  * Decrease reference counter of ipconfig
  */
-void connman_ipconfig_unref(struct connman_ipconfig *ipconfig)
+void __connman_ipconfig_unref_debug(struct connman_ipconfig *ipconfig,
+				const char *file, int line, const char *caller)
 {
 	if (ipconfig == NULL)
 		return;
 
-	DBG("ipconfig %p refcount %d", ipconfig, ipconfig->refcount - 1);
+	DBG("%p ref %d by %s:%d:%s()", ipconfig, ipconfig->refcount - 1,
+		file, line, caller);
 
 	if (__sync_fetch_and_sub(&ipconfig->refcount, 1) != 1)
 		return;
@@ -1333,10 +1346,10 @@ void connman_ipconfig_unref(struct connman_ipconfig *ipconfig)
 	if (__connman_ipconfig_disable(ipconfig) < 0)
 		ipconfig_list = g_list_remove(ipconfig_list, ipconfig);
 
-	connman_ipconfig_set_ops(ipconfig, NULL);
+	__connman_ipconfig_set_ops(ipconfig, NULL);
 
-	if (ipconfig->origin != NULL) {
-		connman_ipconfig_unref(ipconfig->origin);
+	if (ipconfig->origin != NULL && ipconfig->origin != ipconfig) {
+		__connman_ipconfig_unref(ipconfig->origin);
 		ipconfig->origin = NULL;
 	}
 
@@ -1352,7 +1365,7 @@ void connman_ipconfig_unref(struct connman_ipconfig *ipconfig)
  *
  * Get private data pointer
  */
-void *connman_ipconfig_get_data(struct connman_ipconfig *ipconfig)
+void *__connman_ipconfig_get_data(struct connman_ipconfig *ipconfig)
 {
 	if (ipconfig == NULL)
 		return NULL;
@@ -1367,7 +1380,7 @@ void *connman_ipconfig_get_data(struct connman_ipconfig *ipconfig)
  *
  * Set private data pointer
  */
-void connman_ipconfig_set_data(struct connman_ipconfig *ipconfig, void *data)
+void __connman_ipconfig_set_data(struct connman_ipconfig *ipconfig, void *data)
 {
 	ipconfig->ops_data = data;
 }
@@ -1378,7 +1391,7 @@ void connman_ipconfig_set_data(struct connman_ipconfig *ipconfig, void *data)
  *
  * Get interface index
  */
-int connman_ipconfig_get_index(struct connman_ipconfig *ipconfig)
+int __connman_ipconfig_get_index(struct connman_ipconfig *ipconfig)
 {
 	if (ipconfig == NULL)
 		return -1;
@@ -1395,7 +1408,7 @@ int connman_ipconfig_get_index(struct connman_ipconfig *ipconfig)
  *
  * Get interface name
  */
-const char *connman_ipconfig_get_ifname(struct connman_ipconfig *ipconfig)
+const char *__connman_ipconfig_get_ifname(struct connman_ipconfig *ipconfig)
 {
 	struct connman_ipdevice *ipdevice;
 
@@ -1420,7 +1433,7 @@ const char *connman_ipconfig_get_ifname(struct connman_ipconfig *ipconfig)
  *
  * Set the operation callbacks
  */
-void connman_ipconfig_set_ops(struct connman_ipconfig *ipconfig,
+void __connman_ipconfig_set_ops(struct connman_ipconfig *ipconfig,
 				const struct connman_ipconfig_ops *ops)
 {
 	ipconfig->ops = ops;
@@ -1433,7 +1446,7 @@ void connman_ipconfig_set_ops(struct connman_ipconfig *ipconfig,
  *
  * Set the configuration method
  */
-int connman_ipconfig_set_method(struct connman_ipconfig *ipconfig,
+int __connman_ipconfig_set_method(struct connman_ipconfig *ipconfig,
 					enum connman_ipconfig_method method)
 {
 	ipconfig->method = method;
@@ -1456,8 +1469,8 @@ int __connman_ipconfig_address_add(struct connman_ipconfig *ipconfig)
 	switch (ipconfig->method) {
 	case CONNMAN_IPCONFIG_METHOD_UNKNOWN:
 	case CONNMAN_IPCONFIG_METHOD_OFF:
-	case CONNMAN_IPCONFIG_METHOD_AUTO:
 		break;
+	case CONNMAN_IPCONFIG_METHOD_AUTO:
 	case CONNMAN_IPCONFIG_METHOD_FIXED:
 	case CONNMAN_IPCONFIG_METHOD_DHCP:
 	case CONNMAN_IPCONFIG_METHOD_MANUAL:
@@ -1486,8 +1499,8 @@ int __connman_ipconfig_address_remove(struct connman_ipconfig *ipconfig)
 	switch (ipconfig->method) {
 	case CONNMAN_IPCONFIG_METHOD_UNKNOWN:
 	case CONNMAN_IPCONFIG_METHOD_OFF:
-	case CONNMAN_IPCONFIG_METHOD_AUTO:
 		break;
+	case CONNMAN_IPCONFIG_METHOD_AUTO:
 	case CONNMAN_IPCONFIG_METHOD_FIXED:
 	case CONNMAN_IPCONFIG_METHOD_DHCP:
 	case CONNMAN_IPCONFIG_METHOD_MANUAL:
@@ -1514,8 +1527,8 @@ int __connman_ipconfig_address_unset(struct connman_ipconfig *ipconfig)
 	switch (ipconfig->method) {
 	case CONNMAN_IPCONFIG_METHOD_UNKNOWN:
 	case CONNMAN_IPCONFIG_METHOD_OFF:
-	case CONNMAN_IPCONFIG_METHOD_AUTO:
 		break;
+	case CONNMAN_IPCONFIG_METHOD_AUTO:
 	case CONNMAN_IPCONFIG_METHOD_FIXED:
 	case CONNMAN_IPCONFIG_METHOD_DHCP:
 	case CONNMAN_IPCONFIG_METHOD_MANUAL:
@@ -1669,6 +1682,8 @@ int __connman_ipconfig_enable(struct connman_ipconfig *ipconfig)
 	} else
 		return -EINVAL;
 
+	ipconfig->enabled = TRUE;
+
 	if (type == CONNMAN_IPCONFIG_TYPE_IPV4 &&
 					ipdevice->config_ipv4 != NULL) {
 		ipconfig_list = g_list_remove(ipconfig_list,
@@ -1676,7 +1691,7 @@ int __connman_ipconfig_enable(struct connman_ipconfig *ipconfig)
 
 		connman_ipaddress_clear(ipdevice->config_ipv4->system);
 
-		connman_ipconfig_unref(ipdevice->config_ipv4);
+		__connman_ipconfig_unref(ipdevice->config_ipv4);
 	}
 
 	if (type == CONNMAN_IPCONFIG_TYPE_IPV6 &&
@@ -1686,13 +1701,13 @@ int __connman_ipconfig_enable(struct connman_ipconfig *ipconfig)
 
 		connman_ipaddress_clear(ipdevice->config_ipv6->system);
 
-		connman_ipconfig_unref(ipdevice->config_ipv6);
+		__connman_ipconfig_unref(ipdevice->config_ipv6);
 	}
 
 	if (type == CONNMAN_IPCONFIG_TYPE_IPV4)
-		ipdevice->config_ipv4 = connman_ipconfig_ref(ipconfig);
+		ipdevice->config_ipv4 = __connman_ipconfig_ref(ipconfig);
 	else if (type == CONNMAN_IPCONFIG_TYPE_IPV6)
-		ipdevice->config_ipv6 = connman_ipconfig_ref(ipconfig);
+		ipdevice->config_ipv6 = __connman_ipconfig_ref(ipconfig);
 
 	ipconfig_list = g_list_append(ipconfig_list, ipconfig);
 
@@ -1737,11 +1752,13 @@ int __connman_ipconfig_disable(struct connman_ipconfig *ipconfig)
 	if (ipdevice->config_ipv4 == NULL && ipdevice->config_ipv6 == NULL)
 		return -EINVAL;
 
+	ipconfig->enabled = FALSE;
+
 	if (ipdevice->config_ipv4 == ipconfig) {
 		ipconfig_list = g_list_remove(ipconfig_list, ipconfig);
 
 		connman_ipaddress_clear(ipdevice->config_ipv4->system);
-		connman_ipconfig_unref(ipdevice->config_ipv4);
+		__connman_ipconfig_unref(ipdevice->config_ipv4);
 		ipdevice->config_ipv4 = NULL;
 		return 0;
 	}
@@ -1754,7 +1771,7 @@ int __connman_ipconfig_disable(struct connman_ipconfig *ipconfig)
 			disable_ipv6(ipdevice->config_ipv6);
 
 		connman_ipaddress_clear(ipdevice->config_ipv6->system);
-		connman_ipconfig_unref(ipdevice->config_ipv6);
+		__connman_ipconfig_unref(ipdevice->config_ipv6);
 		ipdevice->config_ipv6 = NULL;
 		return 0;
 	}
@@ -2012,7 +2029,7 @@ int __connman_ipconfig_set_config(struct connman_ipconfig *ipconfig,
 	dbus_message_iter_recurse(array, &dict);
 
 	while (dbus_message_iter_get_arg_type(&dict) == DBUS_TYPE_DICT_ENTRY) {
-		DBusMessageIter entry;
+		DBusMessageIter entry, value;
 		const char *key;
 		int type;
 
@@ -2024,7 +2041,12 @@ int __connman_ipconfig_set_config(struct connman_ipconfig *ipconfig,
 		dbus_message_iter_get_basic(&entry, &key);
 		dbus_message_iter_next(&entry);
 
-		type = dbus_message_iter_get_arg_type(&entry);
+		if (dbus_message_iter_get_arg_type(&entry) != DBUS_TYPE_VARIANT)
+			return -EINVAL;
+
+		dbus_message_iter_recurse(&entry, &value);
+
+		type = dbus_message_iter_get_arg_type(&value);
 
 		if (g_str_equal(key, "Method") == TRUE) {
 			const char *str;
@@ -2032,41 +2054,41 @@ int __connman_ipconfig_set_config(struct connman_ipconfig *ipconfig,
 			if (type != DBUS_TYPE_STRING)
 				return -EINVAL;
 
-			dbus_message_iter_get_basic(&entry, &str);
+			dbus_message_iter_get_basic(&value, &str);
 			method = __connman_ipconfig_string2method(str);
 		} else if (g_str_equal(key, "Address") == TRUE) {
 			if (type != DBUS_TYPE_STRING)
 				return -EINVAL;
 
-			dbus_message_iter_get_basic(&entry, &address);
+			dbus_message_iter_get_basic(&value, &address);
 		} else if (g_str_equal(key, "PrefixLength") == TRUE) {
 			if (type != DBUS_TYPE_STRING)
 				return -EINVAL;
 
-			dbus_message_iter_get_basic(&entry,
+			dbus_message_iter_get_basic(&value,
 							&prefix_length_string);
 
 			prefix_length = atoi(prefix_length_string);
 			if (prefix_length < 0 || prefix_length > 128)
 				return -EINVAL;
-
 		} else if (g_str_equal(key, "Netmask") == TRUE) {
 			if (type != DBUS_TYPE_STRING)
 				return -EINVAL;
 
-			dbus_message_iter_get_basic(&entry, &netmask);
+			dbus_message_iter_get_basic(&value, &netmask);
 		} else if (g_str_equal(key, "Gateway") == TRUE) {
 			if (type != DBUS_TYPE_STRING)
 				return -EINVAL;
 
-			dbus_message_iter_get_basic(&entry, &gateway);
+			dbus_message_iter_get_basic(&value, &gateway);
 		} else if (g_str_equal(key, "Privacy") == TRUE) {
 			if (type != DBUS_TYPE_STRING)
 				return -EINVAL;
 
-			dbus_message_iter_get_basic(&entry, &privacy_string);
+			dbus_message_iter_get_basic(&value, &privacy_string);
 			privacy = string2privacy(privacy_string);
 		}
+
 		dbus_message_iter_next(&dict);
 	}
 
@@ -2161,10 +2183,17 @@ int __connman_ipconfig_load(struct connman_ipconfig *ipconfig,
 	key = g_strdup_printf("%smethod", prefix);
 	method = g_key_file_get_string(keyfile, identifier, key, NULL);
 	if (method == NULL) {
-		if (ipconfig->type == CONNMAN_IPCONFIG_TYPE_IPV4)
+		switch (ipconfig->type) {
+		case CONNMAN_IPCONFIG_TYPE_IPV4:
 			ipconfig->method = CONNMAN_IPCONFIG_METHOD_DHCP;
-		else
+			break;
+		case CONNMAN_IPCONFIG_TYPE_IPV6:
+			ipconfig->method = CONNMAN_IPCONFIG_METHOD_AUTO;
+			break;
+		case CONNMAN_IPCONFIG_TYPE_UNKNOWN:
 			ipconfig->method = CONNMAN_IPCONFIG_METHOD_OFF;
+			break;
+		}
 	} else
 		ipconfig->method = __connman_ipconfig_string2method(method);
 
@@ -2181,9 +2210,6 @@ int __connman_ipconfig_load(struct connman_ipconfig *ipconfig,
 			ipconfig->ipv6_privacy_config = string2privacy(privacy);
 			g_free(pprefix);
 			g_free(privacy);
-
-			__connman_ipconfig_enable(ipconfig);
-			enable_ipv6(ipconfig);
 		}
 	}
 
