@@ -32,6 +32,8 @@ static DBusConnection *connection = NULL;
 static GSList *notifier_list = NULL;
 static GHashTable *service_hash = NULL;
 
+static const char *notifier_state;
+
 static gint compare_priority(gconstpointer a, gconstpointer b)
 {
 	const struct connman_notifier *notifier1 = a;
@@ -73,142 +75,68 @@ void connman_notifier_unregister(struct connman_notifier *notifier)
 
 #define MAX_TECHNOLOGIES 10
 
-static volatile int registered[MAX_TECHNOLOGIES];
-static volatile int enabled[MAX_TECHNOLOGIES];
-static volatile int connected[MAX_TECHNOLOGIES];
+static int connected[MAX_TECHNOLOGIES];
+static int online[MAX_TECHNOLOGIES];
 
-void __connman_notifier_list_registered(DBusMessageIter *iter, void *user_data)
+static connman_bool_t notifier_is_online(void)
 {
-	int i;
+	unsigned int i;
 
 	__sync_synchronize();
 	for (i = 0; i < MAX_TECHNOLOGIES; i++) {
-		const char *type = __connman_service_type2string(i);
-
-		if (type == NULL)
-			continue;
-
-		if (registered[i] > 0)
-			dbus_message_iter_append_basic(iter,
-						DBUS_TYPE_STRING, &type);
+		if (online[i] > 0)
+			return TRUE;
 	}
+
+	return FALSE;
 }
 
-void __connman_notifier_list_enabled(DBusMessageIter *iter, void *user_data)
+connman_bool_t __connman_notifier_is_connected(void)
 {
-	int i;
-
-	__sync_synchronize();
-	for (i = 0; i < MAX_TECHNOLOGIES; i++) {
-		const char *type = __connman_service_type2string(i);
-
-		if (type == NULL)
-			continue;
-
-		if (enabled[i] > 0)
-			dbus_message_iter_append_basic(iter,
-						DBUS_TYPE_STRING, &type);
-	}
-}
-
-void __connman_notifier_list_connected(DBusMessageIter *iter, void *user_data)
-{
-	int i;
-
-	__sync_synchronize();
-	for (i = 0; i < MAX_TECHNOLOGIES; i++) {
-		const char *type = __connman_service_type2string(i);
-
-		if (type == NULL)
-			continue;
-
-		if (connected[i] > 0)
-			dbus_message_iter_append_basic(iter,
-						DBUS_TYPE_STRING, &type);
-	}
-}
-
-static void technology_registered(enum connman_service_type type,
-						connman_bool_t registered)
-{
-	DBG("type %d registered %d", type, registered);
-
-	connman_dbus_property_changed_array(CONNMAN_MANAGER_PATH,
-		CONNMAN_MANAGER_INTERFACE, "AvailableTechnologies",
-		DBUS_TYPE_STRING, __connman_notifier_list_registered, NULL);
-}
-
-static void technology_enabled(enum connman_service_type type,
-						connman_bool_t enabled)
-{
-	GSList *list;
-
-	DBG("type %d enabled %d", type, enabled);
-
-	connman_dbus_property_changed_array(CONNMAN_MANAGER_PATH,
-		CONNMAN_MANAGER_INTERFACE, "EnabledTechnologies",
-		DBUS_TYPE_STRING, __connman_notifier_list_enabled, NULL);
-
-	for (list = notifier_list; list; list = list->next) {
-		struct connman_notifier *notifier = list->data;
-
-		if (notifier->service_enabled)
-			notifier->service_enabled(type, enabled);
-	}
-}
-
-unsigned int __connman_notifier_count_connected(void)
-{
-	unsigned int i, count = 0;
+	unsigned int i;
 
 	__sync_synchronize();
 	for (i = 0; i < MAX_TECHNOLOGIES; i++) {
 		if (connected[i] > 0)
-			count++;
+			return TRUE;
 	}
 
-	return count;
+	return FALSE;
+}
+
+static const char *evaluate_notifier_state(void)
+{
+	if (notifier_is_online() == TRUE)
+		return "online";
+
+	if (__connman_notifier_is_connected() == TRUE)
+		return "ready";
+
+	if ( __connman_technology_get_offlinemode() == TRUE)
+		return "offline";
+
+	return "idle";
 }
 
 const char *__connman_notifier_get_state(void)
 {
-	unsigned int count = __connman_notifier_count_connected();
-
-	if (count > 0)
-		return "online";
-
-	return "offline";
+	return notifier_state;
 }
 
-static void state_changed(connman_bool_t connected)
+static void state_changed(void)
 {
-	unsigned int count = __connman_notifier_count_connected();
-	char *state = "offline";
-	DBusMessage *signal;
+	const char *state;
 
-	if (count > 1)
+	state = evaluate_notifier_state();
+
+	if (g_strcmp0(state, notifier_state) == 0)
 		return;
 
-	if (count == 1) {
-		if (connected == FALSE)
-			return;
-
-		state = "online";
-	}
+	notifier_state = state;
 
 	connman_dbus_property_changed_basic(CONNMAN_MANAGER_PATH,
 				CONNMAN_MANAGER_INTERFACE, "State",
-						DBUS_TYPE_STRING, &state);
-
-	signal = dbus_message_new_signal(CONNMAN_MANAGER_PATH,
-				CONNMAN_MANAGER_INTERFACE, "StateChanged");
-	if (signal == NULL)
-		return;
-
-	dbus_message_append_args(signal, DBUS_TYPE_STRING, &state,
-							DBUS_TYPE_INVALID);
-
-	g_dbus_send_message(connection, signal);
+						DBUS_TYPE_STRING, &notifier_state);
 }
 
 static void technology_connected(enum connman_service_type type,
@@ -216,119 +144,8 @@ static void technology_connected(enum connman_service_type type,
 {
 	DBG("type %d connected %d", type, connected);
 
-	connman_dbus_property_changed_array(CONNMAN_MANAGER_PATH,
-		CONNMAN_MANAGER_INTERFACE, "ConnectedTechnologies",
-		DBUS_TYPE_STRING, __connman_notifier_list_connected, NULL);
-
-	state_changed(connected);
-}
-
-void __connman_notifier_register(enum connman_service_type type)
-{
-	DBG("type %d", type);
-
-	switch (type) {
-	case CONNMAN_SERVICE_TYPE_UNKNOWN:
-	case CONNMAN_SERVICE_TYPE_SYSTEM:
-	case CONNMAN_SERVICE_TYPE_GPS:
-	case CONNMAN_SERVICE_TYPE_VPN:
-	case CONNMAN_SERVICE_TYPE_GADGET:
-		return;
-	case CONNMAN_SERVICE_TYPE_ETHERNET:
-	case CONNMAN_SERVICE_TYPE_WIFI:
-	case CONNMAN_SERVICE_TYPE_WIMAX:
-	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
-	case CONNMAN_SERVICE_TYPE_CELLULAR:
-		break;
-	}
-
-	if (__sync_fetch_and_add(&registered[type], 1) == 0)
-		technology_registered(type, TRUE);
-}
-
-void __connman_notifier_unregister(enum connman_service_type type)
-{
-	DBG("type %d", type);
-
-	__sync_synchronize();
-	if (registered[type] == 0) {
-		connman_error("notifier unregister underflow");
-		return;
-	}
-
-	switch (type) {
-	case CONNMAN_SERVICE_TYPE_UNKNOWN:
-	case CONNMAN_SERVICE_TYPE_SYSTEM:
-	case CONNMAN_SERVICE_TYPE_GPS:
-	case CONNMAN_SERVICE_TYPE_VPN:
-	case CONNMAN_SERVICE_TYPE_GADGET:
-		return;
-	case CONNMAN_SERVICE_TYPE_ETHERNET:
-	case CONNMAN_SERVICE_TYPE_WIFI:
-	case CONNMAN_SERVICE_TYPE_WIMAX:
-	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
-	case CONNMAN_SERVICE_TYPE_CELLULAR:
-		break;
-	}
-
-	if (__sync_fetch_and_sub(&registered[type], 1) != 1)
-		return;
-
-	technology_registered(type, FALSE);
-}
-
-void __connman_notifier_enable(enum connman_service_type type)
-{
-	DBG("type %d", type);
-
-	switch (type) {
-	case CONNMAN_SERVICE_TYPE_UNKNOWN:
-	case CONNMAN_SERVICE_TYPE_SYSTEM:
-	case CONNMAN_SERVICE_TYPE_GPS:
-	case CONNMAN_SERVICE_TYPE_VPN:
-	case CONNMAN_SERVICE_TYPE_GADGET:
-		return;
-	case CONNMAN_SERVICE_TYPE_ETHERNET:
-	case CONNMAN_SERVICE_TYPE_WIFI:
-	case CONNMAN_SERVICE_TYPE_WIMAX:
-	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
-	case CONNMAN_SERVICE_TYPE_CELLULAR:
-		break;
-	}
-
-	if (__sync_fetch_and_add(&enabled[type], 1) == 0)
-		technology_enabled(type, TRUE);
-}
-
-void __connman_notifier_disable(enum connman_service_type type)
-{
-	DBG("type %d", type);
-
-	__sync_synchronize();
-	if (enabled[type] == 0) {
-		connman_error("notifier disable underflow");
-		return;
-	}
-
-	switch (type) {
-	case CONNMAN_SERVICE_TYPE_UNKNOWN:
-	case CONNMAN_SERVICE_TYPE_SYSTEM:
-	case CONNMAN_SERVICE_TYPE_GPS:
-	case CONNMAN_SERVICE_TYPE_VPN:
-	case CONNMAN_SERVICE_TYPE_GADGET:
-		return;
-	case CONNMAN_SERVICE_TYPE_ETHERNET:
-	case CONNMAN_SERVICE_TYPE_WIFI:
-	case CONNMAN_SERVICE_TYPE_WIMAX:
-	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
-	case CONNMAN_SERVICE_TYPE_CELLULAR:
-		break;
-	}
-
-	if (__sync_fetch_and_sub(&enabled[type], 1) != 1)
-		return;
-
-	technology_enabled(type, FALSE);
+	__connman_technology_set_connected(type, connected);
+	state_changed();
 }
 
 void __connman_notifier_connect(enum connman_service_type type)
@@ -354,7 +171,16 @@ void __connman_notifier_connect(enum connman_service_type type)
 		technology_connected(type, TRUE);
 }
 
-void __connman_notifier_disconnect(enum connman_service_type type)
+void __connman_notifier_online(enum connman_service_type type)
+{
+	DBG("type %d", type);
+
+	if (__sync_fetch_and_add(&online[type], 1) == 0)
+		state_changed();
+}
+
+void __connman_notifier_disconnect(enum connman_service_type type,
+					enum connman_service_state old_state)
 {
 	DBG("type %d", type);
 
@@ -379,6 +205,9 @@ void __connman_notifier_disconnect(enum connman_service_type type)
 		break;
 	}
 
+	if (old_state == CONNMAN_SERVICE_STATE_ONLINE)
+		__sync_fetch_and_sub(&online[type], 1);
+
 	if (__sync_fetch_and_sub(&connected[type], 1) != 1)
 		return;
 
@@ -401,15 +230,10 @@ static void technology_default(enum connman_service_type type)
 void __connman_notifier_default_changed(struct connman_service *service)
 {
 	enum connman_service_type type = connman_service_get_type(service);
-	char *interface;
 	GSList *list;
 
 	technology_default(type);
 
-	interface = connman_service_get_interface(service);
-	__connman_tethering_update_interface(interface);
-	g_free(interface);
-
 	for (list = notifier_list; list; list = list->next) {
 		struct connman_notifier *notifier = list->data;
 
@@ -482,6 +306,7 @@ void __connman_notifier_offlinemode(connman_bool_t enabled)
 	DBG("enabled %d", enabled);
 
 	offlinemode_changed(enabled);
+	state_changed();
 
 	for (list = notifier_list; list; list = list->next) {
 		struct connman_notifier *notifier = list->data;
@@ -563,54 +388,6 @@ void __connman_notifier_ipconfig_changed(struct connman_service *service,
 	}
 }
 
-static connman_bool_t technology_supported(enum connman_service_type type)
-{
-	switch (type) {
-	case CONNMAN_SERVICE_TYPE_UNKNOWN:
-	case CONNMAN_SERVICE_TYPE_SYSTEM:
-	case CONNMAN_SERVICE_TYPE_GPS:
-	case CONNMAN_SERVICE_TYPE_VPN:
-	case CONNMAN_SERVICE_TYPE_GADGET:
-		return FALSE;
-	case CONNMAN_SERVICE_TYPE_ETHERNET:
-	case CONNMAN_SERVICE_TYPE_WIFI:
-	case CONNMAN_SERVICE_TYPE_WIMAX:
-	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
-	case CONNMAN_SERVICE_TYPE_CELLULAR:
-		break;
-	}
-
-	return TRUE;
-}
-
-connman_bool_t __connman_notifier_is_registered(enum connman_service_type type)
-{
-	DBG("type %d", type);
-
-	if (technology_supported(type) == FALSE)
-		return FALSE;
-
-	__sync_synchronize();
-	if (registered[type] > 0)
-		return TRUE;
-
-	return FALSE;
-}
-
-connman_bool_t __connman_notifier_is_enabled(enum connman_service_type type)
-{
-	DBG("type %d", type);
-
-	if (technology_supported(type) == FALSE)
-		return FALSE;
-
-	__sync_synchronize();
-	if (enabled[type] > 0)
-		return TRUE;
-
-	return FALSE;
-}
-
 int __connman_notifier_init(void)
 {
 	DBG("");
@@ -620,6 +397,7 @@ int __connman_notifier_init(void)
 	service_hash = g_hash_table_new_full(g_direct_hash, g_direct_equal,
 						NULL, NULL);
 
+	notifier_state = evaluate_notifier_state();
 
 	return 0;
 }
